<?php
/**
 * ownCloud
 *
 * @author Artur Neumann <artur@jankaritech.com>
 * @copyright Copyright (c) 2017 Artur Neumann artur@jankaritech.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU AFFERO GENERAL PUBLIC LICENSE for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

use Behat\Behat\Context\Context;
use Behat\Behat\Hook\Scope\BeforeScenarioScope;
use Behat\Gherkin\Node\TableNode;
use Behat\MinkExtension\Context\RawMinkContext;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\GuzzleException;
use Page\FileFirewallPage;
use PHPUnit\Framework\Assert;
use Psr\Http\Message\ResponseInterface;
use TestHelpers\DownloadHelper;
use TestHelpers\Firewall\FirewallHelper;
use TestHelpers\HttpRequestHelper;
use TestHelpers\IpHelper;
use TestHelpers\MoveCopyHelper;
use TestHelpers\OcsApiHelper;
use TestHelpers\SetupHelper;
use TestHelpers\TagsHelper;
use TestHelpers\UploadHelper;
use TestHelpers\WebDavHelper;

require_once 'bootstrap.php';

/**
 * WebUI File Firewall context.
 */
class WebUIFileFirewallContext extends RawMinkContext implements Context {
	private $davPathVersionToUse = 2;
	private $chunkingToUse = null;

	/**
	 * list of tags that were created during the session
	 *
	 * @var array
	 */
	private $createdTags = [];

	/**
	 *
	 * @var FeatureContext
	 */
	private $featureContext;

	/**
	 *
	 * @var IpContext
	 */
	private $ipContext;

	/**
	 *
	 * @var WebUIGeneralContext
	 */
	private $webUIGeneralContext;

	/**
	 *
	 * @var PublicWebDavContext
	 */
	private $publicWebDavContext;

	/**
	 * @var FileFirewallPage
	 */
	private $fileFirewallPage;

	/**
	 * @var string
	 */
	private $tmpDir;

	private $strForNormalFileName
		= 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890';

	private $forbiddenMessage
		= 'Access to this resource has been forbidden by a file firewall rule.';
	/**
	 * @var array
	 */
	private $uploadedFiles = [];
	/**
	 * @var array
	 */
	private $createdFiles = [];

	/**
	 * @var boolean|null
	 */
	private static $appHadToBeEnabled = null;

	/**
	 *
	 * @param FileFirewallPage $fileFirewallPage
	 */
	public function __construct(FileFirewallPage $fileFirewallPage) {
		$this->fileFirewallPage = $fileFirewallPage;
	}

	/**
	 * @When the user browses to the File Firewall settings page
	 * @Given the user has browsed to the File Firewall settings page
	 *
	 * @return void
	 */
	public function theUserBrowsesToTheFileFirewallSettingsPage() {
		$this->fileFirewallPage->open();
		$this->fileFirewallPage->waitForAjaxCallsToStartAndFinish($this->getSession());
	}

	/**
	 * @When the user creates an empty firewall rule group named :name
	 * @Given the user has created an empty firewall rule group named :name
	 *
	 * @param string $name
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserCreatesAnEmptyFirewallRuleGroupNamed($name) {
		$fileFirewallGroup = $this->fileFirewallPage->findGroupByNumber(1);
		if (!$fileFirewallGroup->isEmpty()) {
			$this->theUserCreatesAFirewallRuleGroupNamed($name);
		} else {
			$fileFirewallGroup->setName($name);
		}
	}

	/**
	 * @When the user creates a firewall rule group named :name
	 * @Given the user has created a firewall rule group named :name
	 *
	 * @param string $name
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserCreatesAFirewallRuleGroupNamed($name) {
		$fileFirewallGroup = $this->fileFirewallPage->addGroup();
		$fileFirewallGroup->setName($name);
	}

	/**
	 * @When the user places a rule with the filter :filter, the operator :operator, for the/a :networkScope :ipAddressFamily address subnet and the CIDR :cidr in the group named :groupName
	 * @Given the user has placed a rule with the filter :filter, the operator :operator, for the/a :networkScope :ipAddressFamily address subnet and the CIDR :cidr in the group named :groupName
	 *
	 * @param string $filter
	 * @param string $operator
	 * @param string $networkScope (loopback|routable)
	 * @param string $ipAddressFamily (ipv4|ipv6)
	 * @param string $cidr
	 * @param string $groupName
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserPlacesAnIPv4RangeRuleWithTheFilterTheOperatorForTheAddressSubnetAndTheCIDRInTheGroupNamed(
		$filter, $operator, $networkScope, $ipAddressFamily, $cidr, $groupName
	) {
		$ruleValue
			= IpHelper::ipAddressSubnet($networkScope, $ipAddressFamily, $cidr)
			. '/' . $cidr;
		$this->theUserPlacesARuleWithTheFilterTheOperatorAndTheValueInTheGroupNamed(
			"normal", $filter, $operator, $ruleValue, $groupName
		);
	}

	/**
	 * @When /^the user places a (regex|regular expression|)\s?rule with the filter "([^"]*)", the operator "([^"]*)" and the value "([^"]*)" in the group named "([^"]*)"$/
	 * @Given /^the user has placed a (regex|regular expression|)\s?rule with the filter "([^"]*)", the operator "([^"]*)" and the value "([^"]*)" in the group named "([^"]*)"$/
	 *
	 * @param string $ruleType (regex|regular expression|normal)
	 * @param string $filter
	 * @param string $operator
	 * @param string $value
	 * @param string $groupName
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserPlacesARuleWithTheFilterTheOperatorAndTheValueInTheGroupNamed(
		$ruleType, $filter, $operator, $value, $groupName
	) {
		if ($ruleType === "regex" || $ruleType === "regular expression") {
			$ruleType = "regex";
		} else {
			$ruleType = "normal";
		}
		$value = $this->featureContext->substituteInLineCodes($value);
		$fileFirewallGroup = $this->fileFirewallPage->findGroupByName($groupName);
		$fileFirewallGroup->addRule($filter, $operator, $value, $ruleType);
	}

	/**
	 * @When the user places a rule with the filter :filter, the operator :operator and the value :value in the last group
	 * @Given the user has placed a rule with the filter :filter, the operator :operator and the value :value in the last group
	 *
	 * @param string $filter
	 * @param string $operator
	 * @param string $value
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserPlacesARuleWithTheFilterTheOperatorAndTheValueInTheLastGroup(
		$filter, $operator, $value
	) {
		$value = $this->featureContext->substituteInLineCodes($value);
		$fileFirewallGroup = $this->fileFirewallPage->findGroupByNumber(-1);
		$fileFirewallGroup->addRule($filter, $operator, $value);
	}

	/**
	 * @Given the user deletes the rule no :number in the group named :groupName
	 *
	 * @param int $ruleNo
	 * @param string $groupName
	 *
	 * @return void
	 */
	public function theUserDeletesTheRuleNoInTheGroupNamed($ruleNo, $groupName) {
		$fileFirewallGroup = $this->fileFirewallPage->findGroupByName($groupName);
		$ruleObject = $fileFirewallGroup->getRuleByNo($ruleNo);
		$ruleObject->delete();
	}

	/**
	 * @When /^the user uploads a file (?:named "([^"]*)"|)\s?with the size of "([^"]*)" bytes$/
	 * @Given /^the user has uploaded a file (?:named "([^"]*)"|)\s?with the size of "([^"]*)" bytes$/
	 *
	 * @param string $filename
	 * @param string $bytes
	 *
	 * @return void
	 */
	public function theUserUploadsAFileWithTheSizeOfBytes($filename, $bytes) {
		$this->createFileSpecificSize($filename, $bytes);
		$response = $this->userUploadsAFileTo(
			$this->featureContext->getAdminUsername(),
			$this->tmpDir . "/$filename",
			$filename
		);
		$this->featureContext->setResponse($response);
	}

	/**
	 * @When the user deletes the firewall rule group named :groupName
	 *
	 * @param string $groupName
	 *
	 * @return void
	 * @throws Exception
	 */
	public function theUserDeletesTheFirewallRuleGroupNamed($groupName) {
		$fileFirewallGroup = $this->fileFirewallPage->findGroupByName($groupName);
		$fileFirewallGroup->delete();
	}

	/**
	 * @When the user saves the firewall rules
	 * @Given the user has saved the firewall rules
	 *
	 * @return void
	 */
	public function theUserSavesTheFirewallRules() {
		$this->fileFirewallPage->saveRules($this->getSession());
	}

	/**
	 * checks the dav path version for validity and returns it in an array
	 * so it can be used in a loop. In case of "any" it returns all valid
	 * versions in the array
	 *
	 * @param string $davPathVersion (v1|v2|any)
	 *
	 * @return array
	 * @Transform :davPathVersion
	 */
	public function checkDavPathVersion($davPathVersion) {
		if (!\in_array($davPathVersion, ["v1","v2","any"])) {
			throw new InvalidArgumentException(
				"dav versions can only be 'v1', 'v2' or 'any'"
			);
		}
		if ($davPathVersion === "any") {
			return ["v1","v2"];
		} else {
			return [$davPathVersion];
		}
	}

	/**
	 * checks the "possible" state of a feature step.
	 * e.g. "it should (not|)\s?be possible to ..."
	 *
	 * @param string $possible (not|)
	 *
	 * @return boolean
	 */
	public function checkIfPossible($possible) {
		return ($possible !== "not");
	}

	/**
	 * decides if the proposed dav version and chunking version are
	 * a valid combination.
	 * If no chunkingVersion is specified, then any dav version is valid.
	 * If a chunkingVersion is specified, then it has to match the dav version.
	 * Note: in future the dav and chunking versions might or might not
	 * move together and/or be supported together. So a more complex
	 * matrix could be needed here.
	 *
	 * @param string $davPathVersion
	 * @param string $chunkingVersion
	 *
	 * @return boolean is this a valid combination
	 */
	public function isValidDavChunkingCombination(
		$davPathVersion, $chunkingVersion
	) {
		return (
			($chunkingVersion === 'no') ||
			($davPathVersion === $chunkingVersion)
		);
	}

	/**
	 * checks the chunking version for validity and returns it in an array
	 * so it can be used in a loop. In case of "any" it returns all valid
	 * versions in the array
	 *
	 * @param string $chunkingVersion (no|v1|v2|any)
	 *
	 * @return string[]
	 * @Transform :chunkingVersion
	 */
	public function checkChunkingVersion($chunkingVersion) {
		if (!\in_array($chunkingVersion, ["no","v1","v2","any"])) {
			throw new InvalidArgumentException(
				"chunking versions can only be 'no', 'v1', 'v2' or 'any'"
			);
		}
		if ($chunkingVersion === "any") {
			return ["no", "v1", "v2"];
		} else {
			return [$chunkingVersion];
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible to upload a file with the size of "([^"]*)" bytes (?:and the name "([^"]*)"|)\s?(?:and these http-headers set|)\s?using "(v1|v2|any)" dav path(?: and "?(no|v1|v2|any)"? chunking)?$/
	 *
	 * @param string $possible (not|)
	 * @param string $bytes
	 * @param string $filename
	 * @param string $davPathVersion (v1|v2|any)
	 * @param string $chunkingVersion (no|v1|v2|any)
	 *                                Note: chunking v2 does not work with
	 *                                dav path v1
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 * @throws Exception
	 */
	public function itShouldBePossibleToUploadAFileWithTheSizeOf(
		$possible, $bytes, $filename, $davPathVersion, $chunkingVersion = "no",
		TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);
		$headers = [];
		if ($headersTable !== null) {
			foreach ($headersTable as $header) {
				$headers [$header ["http-header"]] = $header ["value"];
			}
		}
		if ($filename === "") {
			$filename = "filespecificSize.txt";
		}
		$user = $this->featureContext->getAdminUsername();
		$password = $this->featureContext->getAdminPassword();
		$this->createFileSpecificSize($filename, $bytes);
		$davPathVersions = $this->checkDavPathVersion($davPathVersion);
		$chunkingVersions = $this->checkChunkingVersion($chunkingVersion);
		$errorMessage = "using %s dav path and %s chunking";

		foreach ($davPathVersions as $davPathVersion) {
			$this->usingSpecificVersionOfDavPath($davPathVersion);
			foreach ($chunkingVersions as $chunkingVersion) {
				$this->usingSpecificChunking($chunkingVersion);
				if ($this->isValidDavChunkingCombination(
					$davPathVersion, $chunkingVersion
				)
				) {
					$response = $this->userUploadsAFileTo(
						$user, $this->tmpDir . "/$filename", $filename, $headers
					);
					//you will have a 500 Error here if the file already exists
					//on the server
					//see: https://github.com/owncloud/firewall/issues/254
					FirewallHelper::assertHttpResultCorrect(
						$possible,
						$response->getStatusCode(),
						\sprintf($errorMessage, $davPathVersion, $chunkingVersion)
					);
					$this->assertFileShouldOrNotExist($user, $filename, $possible);

					if ($possible) {
						// Attempt to delete the uploaded file so that the next
						// upload will be a true upload (not an overwrite)
						try {
							WebDavHelper::makeDavRequest(
								$this->featureContext->getBaseUrl(),
								$user,
								$password,
								"DELETE",
								$filename,
								[]
							);
							unset($this->uploadedFiles[$user][$filename]);
						} catch (BadResponseException $e) {
							echo "INFORMATION: could not delete file '$filename'" .
								" from server after upload. code: " .
								$e->getResponse()->getStatusCode() . "\n";
						}
					}
				}
			}
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible to upload a file with the size of "([^"]*)" bytes (?:and the name "([^"]*)"|)\s?(?:and these http-headers set|)\s?using "(v1|v2|any)" dav path(?: and "?(no|v1|v2|any)"? chunking)? on the last created public link$/
	 *
	 * @param string $possible (not|)
	 * @param string $bytes
	 * @param string $filename
	 * @param string $davPathVersion (v1|v2|any)
	 * @param string $chunkingVersion (no|v1|v2|any)
	 *                                Note: chunking v2 does not work with
	 *                                dav path v1
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 * @throws Exception
	 */
	public function itShouldBePossibleToUploadAFileWithTheSizeOfOnLastPublicLink(
		$possible, $bytes, $filename, $davPathVersion, $chunkingVersion = "no",
		TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);
		$headers = [];
		if ($headersTable !== null) {
			foreach ($headersTable as $header) {
				$headers [$header ["http-header"]] = $header ["value"];
			}
		}
		if ($filename === "") {
			$filename = "filespecificSize.txt";
		}
		$password = $this->featureContext->getAdminPassword();
		$this->createFileSpecificSize($filename, $bytes);
		$davPathVersions = $this->checkDavPathVersion($davPathVersion);
		$chunkingVersions = $this->checkChunkingVersion($chunkingVersion);
		$errorMessage = "using %s dav path and %s chunking";

		foreach ($davPathVersions as $davPathVersion) {
			$this->usingSpecificVersionOfDavPath($davPathVersion);
			foreach ($chunkingVersions as $chunkingVersion) {
				$this->usingSpecificChunking($chunkingVersion);
				if ($this->isValidDavChunkingCombination(
					$davPathVersion, $chunkingVersion
				)
				) {
					$fileName = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
					$this->createFileSpecificSize($fileName, $bytes);
					$this->publicWebDavContext->publicUploadContent(
						$fileName,
						'',
						\file_get_contents($this->tmpDir . "/$fileName"),
						false,
						$headers
					);
					$response = $this->featureContext->getResponse();

					//you will have a 500 Error here if the file already exists
					//on the server
					//see: https://github.com/owncloud/firewall/issues/254
					FirewallHelper::assertHttpResultCorrect(
						$possible,
						$response->getStatusCode(),
						\sprintf($errorMessage, $davPathVersion, $chunkingVersion)
					);
					$user = $this->featureContext->getLastShareData()->data->uid_owner->__toString();
					$targetName = $this->featureContext->getLastShareData()->data->path->__toString() . '/' . $fileName;
					$this->assertFileShouldOrNotExist($user, $targetName, $possible);

					if ($possible) {
						// Attempt to delete the uploaded file so that the next
						// upload will be a true upload (not an overwrite)
						try {
							WebDavHelper::makeDavRequest(
								$this->featureContext->getBaseUrl(),
								$user,
								$password,
								"DELETE",
								$targetName,
								[]
							);
							unset($this->uploadedFiles[$user][$filename]);
						} catch (BadResponseException $e) {
							echo "INFORMATION: could not delete file '$filename'" .
								" from server after upload. code: " .
								$e->getResponse()->getStatusCode() . "\n";
						}
					}
				}
			}
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible to upload a file with the mime-type "([^"]*)" (?:and these http-headers set|)\s?using "(v1|v2|any)" dav path(?: and "?(no|v1|v2|any)"? chunking)?$/
	 *
	 * @param string $possible (not|)
	 * @param string $mimeType
	 * @param string $davPathVersions (v1|v2|any)
	 * @param string $chunkingVersion (no|v1|v2|any)
	 *                                Note: chunking v2 does not work with
	 *                                dav path v1
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToUploadAFileWithTheMimeType(
		$possible, $mimeType, $davPathVersions, $chunkingVersion = "no",
		TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);

		$headers = [];
		if ($headersTable !== null) {
			foreach ($headersTable as $header) {
				$headers [$header ["http-header"]] = $header ["value"];
			}
		}

		$extensions = $this->getAllExtensionsForMimeType($mimeType);
		$davPathVersions = $this->checkDavPathVersion($davPathVersions);
		$chunkingVersions = $this->checkChunkingVersion($chunkingVersion);
		$errorMessage = "using %s dav path and %s chunking";

		foreach ($extensions as $extension) {
			foreach ($davPathVersions as $davPathVersion) {
				$this->usingSpecificVersionOfDavPath($davPathVersion);
				foreach ($chunkingVersions as $chunkingVersion) {
					$this->usingSpecificChunking($chunkingVersion);
					if ($this->isValidDavChunkingCombination(
						$davPathVersion, $chunkingVersion
					)
					) {
						$user = $this->featureContext->getAdminUsername();
						$fileName = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
						$response = $this->uploadAFileWithTheExtension(
							$extension, $headers, $user, $fileName
						);
						FirewallHelper::assertHttpResultCorrect(
							$possible,
							$response->getStatusCode(),
							\sprintf($errorMessage, $davPathVersion, $chunkingVersion)
						);
						$this->assertFileShouldOrNotExist($user, $fileName . '.' . $extension, $possible);
					}
				}
			}
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible to upload a file with the mime-type "([^"]*)" (?:and these http-headers set|)\s?using "(v1|v2|any)" dav path(?: and "?(no|v1|v2|any)"? chunking)? on the last created public link$/
	 *
	 * @param string $possible (not|)
	 * @param string $mimeType
	 * @param string $davPathVersions (v1|v2|any)
	 * @param string $chunkingVersion (no|v1|v2|any)
	 *                                Note: chunking v2 does not work with
	 *                                dav path v1
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToUploadAFileWithTheMimeTypeOnLastPublicLink(
		$possible, $mimeType, $davPathVersions, $chunkingVersion = "no",
		TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);

		$headers = [];
		if ($headersTable !== null) {
			foreach ($headersTable as $header) {
				$headers [$header ["http-header"]] = $header ["value"];
			}
		}

		$extensions = $this->getAllExtensionsForMimeType($mimeType);
		$davPathVersions = $this->checkDavPathVersion($davPathVersions);
		$chunkingVersions = $this->checkChunkingVersion($chunkingVersion);
		$errorMessage = "using %s dav path and %s chunking";

		foreach ($extensions as $extension) {
			foreach ($davPathVersions as $davPathVersion) {
				$this->usingSpecificVersionOfDavPath($davPathVersion);
				foreach ($chunkingVersions as $chunkingVersion) {
					$this->usingSpecificChunking($chunkingVersion);
					if ($this->isValidDavChunkingCombination(
						$davPathVersion, $chunkingVersion
					)
					) {
						$fileName = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
						$fileName = $fileName . '.' . $extension;
						$this->createFileSpecificSize($fileName, 1024);
						$this->publicWebDavContext->publicUploadContent(
							$fileName,
							'',
							\file_get_contents($this->tmpDir . "/$fileName"),
							false,
							$headers
						);
						$response = $this->featureContext->getResponse();
						FirewallHelper::assertHttpResultCorrect(
							$possible,
							$response->getStatusCode(),
							\sprintf($errorMessage, $davPathVersion, $chunkingVersion)
						);
						$user = $this->featureContext->getLastShareData()->data->uid_owner->__toString();
						$targetName = $this->featureContext->getLastShareData()->data->path->__toString() . '/' . $fileName;
						$this->assertFileShouldOrNotExist($user, $targetName, $possible);
					}
				}
			}
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible to upload a file with the "([^"]*)" user agent using "(v1|v2|any)" dav path$/
	 *
	 * @param string $possible (not|)
	 * @param string $userAgent
	 * @param string $davPathVersion (v1|v2|any)
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToUploadAFileWithTheUserAgent(
		$possible, $userAgent, $davPathVersion
	) {
		$possible = $this->checkIfPossible($possible);

		$davPathVersion = $this->checkDavPathVersion($davPathVersion);
		foreach ($davPathVersion as $version) {
			$this->usingSpecificVersionOfDavPath($version);
			$user = $this->featureContext->getAdminUsername();
			$fileName = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
			$response = $this->uploadAFileWithTheUserAgent(
				$userAgent, $user, $fileName
			);
			FirewallHelper::assertHttpResultCorrect(
				$possible,
				$response->getStatusCode(),
				"using $version dav path"
			);
			$this->assertFileShouldOrNotExist($user, $fileName, $possible);
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible for user "([^"]*)" to download the file "([^"]*)" with the "([^"]*)" user agent using "(v1|v2|any)" dav path$/
	 *
	 * @param string $possible (not|)
	 * @param string $user
	 * @param string $fileName
	 * @param string $userAgent
	 * @param string $davPathVersion (v1|v2|any)
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToDownloadTheFileWithTheUserAgent(
		$possible, $user, $fileName, $userAgent, $davPathVersion
	) {
		$possible = $this->checkIfPossible($possible);

		$davPathVersion = $this->checkDavPathVersion($davPathVersion);
		foreach ($davPathVersion as $version) {
			$this->usingSpecificVersionOfDavPath($version);
			$response = $this->userDownloadsAFile(
				$user,
				$fileName,
				["User-Agent" => $userAgent]
			);
			FirewallHelper::assertHttpResultCorrect(
				$possible,
				$response->getStatusCode(),
				"using $version dav path"
			);
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible for user "([^"]*)" to download the file "([^"]*)" using "(v1|v2|any)" dav path$/
	 *
	 * @param string $possible (not|)
	 * @param string $user
	 * @param string $fileName
	 * @param string $davPathVersion (v1|v2|any)
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToDownloadTheFile(
		$possible, $user, $fileName, $davPathVersion
	) {
		$possible = $this->checkIfPossible($possible);

		$davPathVersion = $this->checkDavPathVersion($davPathVersion);
		foreach ($davPathVersion as $version) {
			$this->usingSpecificVersionOfDavPath($version);
			$response = $this->userDownloadsAFile(
				$user,
				$fileName,
				[]
			);
			FirewallHelper::assertHttpResultCorrect(
				$possible,
				$response->getStatusCode(),
				"using $version dav path"
			);
		}
	}

	/**
	 * @When /^the public downloads (file|folder) "([^"]*)" through the last created public link$/
	 *
	 * @param string $fileOrFolder (file|folder)
	 * @param string $itemName the file or folder name
	 * @param array $headers headers to send with the request
	 *
	 * @return ResponseInterface
	 */
	public function thePublicDownloadsFileThroughTheLastCreatedPublicLink(
		$fileOrFolder, $itemName, $headers = []
	) {
		$downloadURL = null;
		$shareType = $this->featureContext->getLastShareData()->data->item_type->__toString();
		if ($shareType === $fileOrFolder
			&& \preg_match(
				"/\/?" . \preg_quote($itemName, "/") . "/",
				$this->featureContext->getLastShareData()->data->path
			)
		) {
			$downloadURL = $this->featureContext->getLastShareData()->data->url . "/download";
		} elseif ($shareType === "folder" && $fileOrFolder === "file") {
			$pathInfo = \pathinfo($itemName);
			if (\substr($pathInfo['dirname'], 0, 1) !== "/") {
				$pathInfo['dirname'] = "/" . $pathInfo['dirname'];
			}
			$downloadURL = $this->featureContext->getLastShareData()->data->url .
						   "/download?path=" . $pathInfo ['dirname'] .
						   "&files=" . $pathInfo ['basename'];
		}

		Assert::assertNotNull($downloadURL);

		$response = HttpRequestHelper::get($downloadURL, null, null, $headers);
		$this->featureContext->setResponse($response);

		return $response;
	}

	/**
	 * @Then the response should indicate that access has been forbidden by a file firewall rule
	 *
	 * @return void
	 */
	public function theResponseShouldIndicateForbiddenByFileFirewallRule() {
		$response = $this->featureContext->getResponse();
		Assert::assertStringContainsString(
			$this->forbiddenMessage,
			$response->getBody()->getContents()
		);
	}

	/**
	 * @Then /^it should (not|)\s?be possible to download the (file|folder) "([^"]*)" through the last created public link$/
	 *
	 * @param string $possible (not|)
	 * @param string $fileOrFolder (file|folder)
	 * @param string $itemName the file or folder name
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToDownloadTheFileThroughTheLastCreatedPublicLink(
		$possible, $fileOrFolder, $itemName
	) {
		$possible = $this->checkIfPossible($possible);

		$response = $this->thePublicDownloadsFileThroughTheLastCreatedPublicLink(
			$fileOrFolder, $itemName
		);

		if ($possible) {
			Assert::assertEquals(200, $response->getStatusCode());
			Assert::assertNotContains(
				$this->forbiddenMessage,
				$response->getBody()->getContents()
			);
		} else {
			Assert::assertEquals(403, $response->getStatusCode());
			Assert::assertStringContainsString(
				$this->forbiddenMessage,
				$response->getBody()->getContents()
			);
		}
	}

	/**
	 * @Then /^it should (not|)\s?be possible for user "?([^"]*)" to view the preview of "?([^"]*)"?\s?(?:using (?:the )?(old|new|any) preview url)?(?: and these http-headers set|)$/
	 *
	 * @param string $possible (not|)
	 * @param string $user
	 * @param string $fileName
	 * @param string $previewURLversion
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 * @throws Exception
	 */
	public function viewThePreview(
		$possible,
		$user,
		$fileName,
		$previewURLversion = "any",
		TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);
		$fileNameEnc = \rawurlencode($fileName);
		$oldPreviewURL = $this->ipContext->getBaseUrlForSourceIp() .
					  "/index.php/core/preview.png?file=" . $fileNameEnc .
					  "&x=128&y=128&forceIcon=0";
		//after https://github.com/owncloud/core/pull/29319
		//the previews are generated by a dav endpoint
		$newPreviewURL = $this->ipContext->getBaseUrlForSourceIp() .
					  "/remote.php/dav/files/" . $user . "/" . $fileName .
					  "?x=39&y=39&preview=1";
		$options = [
			'auth' => [
				$user,
				$this->featureContext->getUserPassword($user)
			],
			'allow_redirects' => false
		];
		if ($this->ipContext->getSourceIpAddress() !== null) {
			$options['config']
				= [ 'curl' => [ CURLOPT_INTERFACE => $this->ipContext->getSourceIpAddress() ]];
		}
		$guzzleClient = new Client($options);

		$headers = FirewallHelper::createHttpRequestHeadersArray($headersTable);
		$request = HttpRequestHelper::createRequest(
			$oldPreviewURL, "GET", $headers
		);

		try {
			if ($previewURLversion === "any" || $previewURLversion === "old") {
				$response = $guzzleClient->send($request);
			}

			if (($previewURLversion === "any" && $response->getStatusCode() === 302)
				|| $previewURLversion === "new"
			) {
				//we seem to use the new URL
				$request = HttpRequestHelper::createRequest(
					$newPreviewURL, "GET", $headers
				);
				$response = $guzzleClient->send($request);
			} elseif ($previewURLversion === "old"
				&& $response->getStatusCode() === 302
			) {
				//we had to check the old URL and that returned 302
				//so its not active anymore, nothing left to do
				return;
			}
		} catch (GuzzleException $e) {
			$response = $e->getResponse();
		}

		FirewallHelper::assertHttpResultCorrect(
			$possible,
			$response->getStatusCode(), "view preview of file $fileName"
		);
	}

	/**
	 * wrapper around viewThePreview() just sets the user agent
	 * in the http-header
	 *
	 * @Then /^it should (not|)\s?be possible for user "?([^"]*)" to view the preview of "?([^"]*)"? with the "?([^"]*)"? user agent$/
	 *
	 * @param string $possible (not|)
	 * @param string $user
	 * @param string $fileName
	 * @param string $userAgent
	 *
	 * @return void
	 * @throws Exception
	 */
	public function viewThePreviewWithUserAgent(
		$possible, $user, $fileName, $userAgent
	) {
		$table = new TableNode(
			[
				0 => [
					"http-header",
					"value"
				],
				1 => [
					"User-Agent",
					$userAgent
				]
			]
		);

		$this->viewThePreview($possible, $user, $fileName, "any", $table);
	}

	/**
	 * @Then /^it should (not|)\s?be possible to view the preview of "?([^"]*)"? through the last created public link(?: with these http-headers set|)$/
	 *
	 * @param string $possible (not|)
	 * @param string $fileName
	 * @param TableNode $headersTable table of headers to be set for the upload
	 *                                table headings: must be: |http-header|value|
	 *
	 * @return void
	 */
	public function viewThePreviewThroughPublicLink(
		$possible, $fileName, TableNode $headersTable = null
	) {
		$possible = $this->checkIfPossible($possible);

		$token = \pathinfo($this->featureContext->getLastShareData()->data->url)['filename'];
		$fileName = \rawurlencode($fileName);
		$previewURL = $this->ipContext->getBaseUrlForSourceIp() .
			"/index.php/apps/files_sharing/ajax/publicpreview.php?file=" .
			$fileName . "&x=128&y=128&t=" . $token;

		$options = [];
		if ($this->ipContext->getSourceIpAddress() !== null) {
			$options['config']
				= [ 'curl' => [ CURLOPT_INTERFACE => $this->ipContext->getSourceIpAddress() ]];
		}
		$guzzleClient = new Client($options);
		$request = HttpRequestHelper::createRequest(
			$previewURL,
			"GET",
			FirewallHelper::createHttpRequestHeadersArray($headersTable)
		);
		try {
			$response = $guzzleClient->send($request);
		} catch (GuzzleException $e) {
			$response = $e->getResponse();
		}

		FirewallHelper::assertHttpResultCorrect(
			$possible,
			$response->getStatusCode(),
			"view preview of $fileName through public link"
		);
	}

	/**
	 * wrapper around viewThePreviewThroughPublicLink() just sets the user agent
	 * in the http-header
	 *
	 * @Then /^it should (not|)\s?be possible to view the preview of "?([^"]*)"? through the last created public link with the "?([^"]*)"? user agent$/
	 *
	 * @param string $possible (not|)
	 * @param string $fileName
	 * @param string $userAgent
	 *
	 * @return void
	 */
	public function viewThePreviewThroughPublicLinkWithUserAgent(
		$possible, $fileName, $userAgent
	) {
		$table = new TableNode(
			[
				0 => [
					"http-header",
					"value"
				],
				1 => [
					"User-Agent",
					$userAgent
				]
			]
		);

		$this->viewThePreviewThroughPublicLink($possible, $fileName, $table);
	}

	/**
	 * @Then it should be possible to download the file :file through the last created public link with the :agent user agent
	 *
	 * @param string $file
	 * @param string $agent
	 *
	 * @return void
	 */
	public function itShouldBePossibleToDownloadTheFileThroughTheLastCreatedPublicLinkWithTheUserAgent($file, $agent) {
		$header = ["User-Agent" => $agent];
		$response = $this->thePublicDownloadsFileThroughTheLastCreatedPublicLink('file', $file, $header);
		FirewallHelper::assertHttpResultCorrect(
			true,
			$response->getStatusCode(),
			"download $file through public link"
		);
	}

	/**
	 * @Then it should not be possible to download the file :file through the last created public link with the :agent user agent
	 *
	 * @param string $file
	 * @param string $agent
	 *
	 * @return void
	 */
	public function itShouldNotBePossibleToDownloadTheFileThroughTheLastCreatedPublicLinkWithTheUserAgent($file, $agent) {
		$header = ["User-Agent" => $agent];
		$response = $this->thePublicDownloadsFileThroughTheLastCreatedPublicLink('file', $file, $header);
		FirewallHelper::assertHttpResultCorrect(
			false,
			$response->getStatusCode(),
			"download $file through public link"
		);
	}

	/**
	 * @Then /^it should (not|)\s?be possible to (move|copy) the file "?([^"]*)"?\s?(asynchronously|) to "?([^"]*)"? using "(v1|v2|any)" dav path$/
	 *
	 * @param string $possible
	 * @param string $moveOrCopy
	 * @param string $fromFileName
	 * @param string $asyncOrNot "asynchronously" or empty
	 * @param string $toFileName
	 * @param string $davPathVersion
	 *
	 * @return void
	 * @throws Exception
	 */
	public function itShouldNotBePossibleToMoveTheFileUsingDavPath(
		$possible,
		$moveOrCopy,
		$fromFileName,
		$asyncOrNot,
		$toFileName,
		$davPathVersion
	) {
		$possible = $this->checkIfPossible($possible);
		$headers = [];
		if ($asyncOrNot === "asynchronously") {
			$headers['OC-LazyOps'] = 'true';
		}
		$user = $this->featureContext->getAdminUsername();
		$password = $this->featureContext->getAdminPassword();
		$errorMessage = "using %s dav path";
		$davPathVersion = $this->checkDavPathVersion($davPathVersion);
		foreach ($davPathVersion as $version) {
			$this->usingSpecificVersionOfDavPath($version);
			try {
				if ($moveOrCopy === "copy") {
					$response = MoveCopyHelper::copy(
						$this->featureContext->getBaseUrl(),
						$user,
						$password,
						$fromFileName,
						$toFileName,
						$headers
					);
				} elseif ($moveOrCopy === "move") {
					$response = MoveCopyHelper::move(
						$this->featureContext->getBaseUrl(),
						$user,
						$password,
						$fromFileName,
						$toFileName,
						$headers
					);

					unset($this->uploadedFiles[$user][$fromFileName]);
				} else {
					throw new \InvalidArgumentException(
						'$moveOrCopy has to be "copy" or "move"'
					);
				}
				$this->uploadedFiles[$user][$toFileName] = "";
			} catch (BadResponseException $e) {
				// 4xx and 5xx responses cause an exception
				$response = $e->getResponse();
			}
			FirewallHelper::assertHttpResultCorrect(
				$possible,
				$response->getStatusCode(),
				\sprintf($errorMessage, $version)
			);
			$this->assertFileShouldOrNotExist($user, $toFileName, $possible);
		}
	}

	/**
	 * sets the dav path version to use from human readable format
	 *
	 * @Given /^using "(v1|v2)" dav path$/
	 *
	 * @param string $version (v1|v2)
	 *
	 * @return void
	 */
	public function usingSpecificVersionOfDavPath($version) {
		if ($version === "v1") {
			$this->davPathVersionToUse = 1;
		} elseif ($version === "v2") {
			$this->davPathVersionToUse = 2;
		} else {
			throw new InvalidArgumentException(
				"cannot set dav path version to $version"
			);
		}
	}

	/**
	 * sets the chunking version from human readable format
	 *
	 * @Given /^using "(no|v1|v2)" chunking$/
	 *
	 * @param string $version (no|v1|v2)
	 *
	 * @return void
	 */
	public function usingSpecificChunking($version) {
		if ($version === "v1") {
			$this->chunkingToUse = 1;
		} elseif ($version === "v2") {
			$this->chunkingToUse = 2;
		} elseif ($version === "no") {
			$this->chunkingToUse = null;
		} else {
			throw new InvalidArgumentException(
				"cannot set chunking version to $version"
			);
		}
	}

	/**
	 * uploads a file using the given User Agent
	 *
	 * @param string $userAgent
	 * @param string|null $user
	 * @param string|null $filename
	 *
	 * @return ResponseInterface|NULL|\GuzzleHttp\Message\FutureResponse
	 */
	public function uploadAFileWithTheUserAgent($userAgent, $user = null, $filename = null) {
		if ($user === null) {
			$user = $this->featureContext->getAdminUsername();
		}
		if ($filename === null) {
			$filename = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
		}
		$this->createFileSpecificSize($filename, 1);
		$response = $this->userUploadsAFileTo(
			$user,
			$this->tmpDir . "/$filename",
			$filename,
			["User-Agent" => $userAgent]
		);
		return $response;
	}

	/**
	 * creates and uploads a file with a random name and the given extension
	 *
	 * @param string $extension
	 * @param array $headers HTTP Headers to set
	 * @param string|null $user
	 * @param string|null $filename
	 *
	 * @return ResponseInterface|NULL|\GuzzleHttp\Message\FutureResponse
	 */
	public function uploadAFileWithTheExtension($extension, $headers, $user = null, $filename = null) {
		if ($user === null) {
			$user = $this->featureContext->getAdminUsername();
		}
		if ($filename === null) {
			$filename = \substr(\str_shuffle($this->strForNormalFileName), 0, 16);
		}
		$filename .= "." . $extension;
		$this->createFileSpecificSize($filename, 1024);
		$response = $this->userUploadsAFileTo(
			$user, $this->tmpDir . "/$filename", $filename, $headers
		);
		return $response;
	}

	/**
	 * @When user :user uploads file :source to :destination
	 *
	 * @param string $user
	 * @param string $source
	 * @param string $destination
	 * @param array $headers
	 *
	 * @return ResponseInterface|NULL|\GuzzleHttp\Message\FutureResponse
	 */
	public function userUploadsAFileTo(
		$user, $source, $destination, $headers = []
	) {
		if ($this->chunkingToUse !== null) {
			$noOfChunks = 3; //for the sake of tests use 3 chunks
		} else {
			$noOfChunks = 1;
		}
		try {
			$response = UploadHelper::upload(
				$this->featureContext->getBaseUrl(),
				$this->featureContext->getActualUsername($user),
				$this->featureContext->getUserPassword($user),
				$source,
				$destination,
				$headers,
				$this->davPathVersionToUse,
				$this->chunkingToUse,
				$noOfChunks
			);
			$this->uploadedFiles[$user][$destination] = "";
		} catch (BadResponseException $e) {
			// 4xx and 5xx responses cause an exception
			$response = $e->getResponse();
		}
		return $response;
	}

	/**
	 * @When /^the user "?([^"]*)" tags the (?:file|folder) "(.*)" with these tags/
	 *
	 * @param string $user
	 * @param string $fileName
	 * @param TableNode $table Table of tags, no heading
	 *
	 * @return void
	 * @throws Exception
	 */
	public function userTagsTheFileWithTheseTags(
		$user, $fileName, TableNode $table
	) {
		foreach ($table->getTable() as $row) {
			$tag = $row[0];
			TagsHelper::tag(
				$this->featureContext->getBaseUrl(),
				$user,
				$this->featureContext->getUserPassword($user),
				$tag,
				$fileName,
				null,
				null,
				2
			);
		}
	}

	/**
	 * set log level of firewall
	 *
	 * @When the user sets the firewall logging level to :logLevel
	 *
	 * @param string $logLevel
	 *
	 * @return void
	 */
	public function theUserSetsTheFirewallLoggingLevelTo($logLevel) {
		$this->fileFirewallPage->setLogLevel($logLevel, $this->getSession());
	}

	/**
	 * download a file using a specific user
	 *
	 * @param string $user
	 * @param string $fileName
	 * @param string[] $headers
	 *
	 * @return ResponseInterface
	 */
	public function userDownloadsAFile($user, $fileName, $headers) {
		try {
			$response = DownloadHelper::download(
				$this->ipContext->getBaseUrlForSourceIp(),
				$user,
				$this->featureContext->getUserPassword($user),
				$fileName,
				$headers,
				$this->davPathVersionToUse,
				$this->ipContext->getSourceIpAddress()
			);
		} catch (BadResponseException $e) {
			// 4xx and 5xx responses cause an exception
			$response = $e->getResponse();
		}
		return $response;
	}

	/**
	 * creates a File with a specific size
	 *
	 * @param string $name
	 * @param int $size
	 *
	 * @return void
	 */
	public function createFileSpecificSize($name, $size) {
		UploadHelper::createFileSpecificSize($this->tmpDir . "/" . "$name", $size);
		$this->createdFiles[$name] = '';
	}

	/**
	 * creates a File with a specific text content
	 *
	 * @param string $name
	 * @param string $text
	 *
	 * @return void
	 */
	public function createFileWithText($name, $text) {
		UploadHelper::createFileWithText($this->tmpDir . "/" . "$name", $text);
		$this->createdFiles[$name] = '';
	}

	/**
	 * Assert the file exist or not for a user based on the condition
	 *
	 * @param string $user
	 * @param string $filename
	 * @param bool $should Boolean value indicating if the file should exist of not
	 *
	 * @return void
	 * @throws Exception
	 */
	public function assertFileShouldOrNotExist($user, $filename, $should) {
		if ($should) {
			$this->featureContext->asFileOrFolderShouldExist($user, "file", $filename);
		} else {
			$this->featureContext->asFileOrFolderShouldNotExist($user, "file", $filename);
		}
	}

	/**
	 * This will run once at the end of a whole suite
	 * of features with scenarios.
	 *
	 * @AfterSuite
	 *
	 * @return void
	 */
	public static function afterFirewallSuite() {
		if (self::$appHadToBeEnabled) {
			SetupHelper::disableApp('firewall');
		}
	}

	/**
	 * general before scenario for all firewall tests.
	 * This will run before EVERY scenario.
	 * It will set the properties for this object.
	 *
	 * @BeforeScenario
	 *
	 * @param BeforeScenarioScope $scope
	 *
	 * @return void
	 */
	public function before(BeforeScenarioScope $scope) {
		// Get the environment
		$environment = $scope->getEnvironment();

		// Get all the contexts you need in this context
		$this->featureContext = $environment->getContext('FeatureContext');
		$this->ipContext = $environment->getContext('IpContext');
		$this->webUIGeneralContext = $environment->getContext('WebUIGeneralContext');
		$this->publicWebDavContext = $environment->getContext('PublicWebDavContext');
		$this->tmpDir = $this->getMinkParameter("show_tmp_dir");

		SetupHelper::init(
			$this->featureContext->getAdminUsername(),
			$this->featureContext->getAdminPassword(),
			$this->featureContext->getBaseUrl(),
			$this->featureContext->getOcPath()
		);

		//only do the enabling once
		if (self::$appHadToBeEnabled === null) {
			self::$appHadToBeEnabled = SetupHelper::enableAppIfNotEnabled(
				'firewall'
			);
		}
		FirewallHelper::deleteAllFirewallRules();
	}

	/**
	 * general after scenario for all firewall tests.
	 * Cleans up created tags, uploaded and created files
	 *
	 * @AfterScenario
	 *
	 * @return void
	 */
	public function after() {
		FirewallHelper::deleteAllFirewallRules();

		$adminUsername = $this->featureContext->getAdminUsername();
		$adminPassword = $this->featureContext->getAdminPassword();

		foreach ($this->createdTags as $tagID) {
			try {
				TagsHelper::deleteTag(
					$this->featureContext->getBaseUrl(),
					$adminUsername,
					$adminPassword,
					$tagID,
					2
				);
			} catch (ClientException  $e) {
				echo "INFORMATION: could not delete tag $tagID. " .
					$e->getResponse()->getStatusCode();
			}
		}
		//delete files that were uploaded to the server
		foreach ($this->uploadedFiles as $user => $files) {
			foreach (\array_keys($files) as $file) {
				$currentTime = \microtime(true);
				$end = $currentTime + (STANDARD_UI_WAIT_TIMEOUT_MILLISEC / 1000);
				while ($currentTime <= $end) {
					try {
						WebDavHelper::makeDavRequest(
							$this->featureContext->getBaseUrl(),
							$user,
							$this->featureContext->getUserPassword($user),
							"DELETE",
							$file,
							[]
						);
						break;
					} catch (BadResponseException $e) {
						// 4xx and 5xx responses cause an exception
						//let the test pass, but display an info
						if ($e->getResponse()->getStatusCode() === 403) {
							echo "INFORMATION: could not delete file '$file'" .
								" from server. " .
								"code: " . $e->getResponse()->getStatusCode() .
								" maybe a firewall rule is still active.\n";
						} else {
							echo "INFORMATION: could not delete file '$file'" .
								" from server\n" . $e->getResponse();
							break;
						}
					}
					\usleep(STANDARD_SLEEP_TIME_MICROSEC);
					$currentTime = \microtime(true);
				}
			}
		}
		//delete tmp files
		foreach (\array_keys($this->createdFiles) as $fileName) {
			if (\unlink($this->tmpDir . "/" . $fileName) === false) {
				echo "INFORMATION: could not delete file '$fileName'";
			}
		}
	}

	/**
	 * returns a list of file extensions that are associated with a mime-type
	 *
	 * @param string $mimeTypeNeedle
	 *
	 * @return string[]
	 */
	private function getAllExtensionsForMimeType($mimeTypeNeedle) {
		$response = OcsApiHelper::sendRequest(
			$this->featureContext->getBaseUrl(),
			$this->featureContext->getAdminUsername(),
			$this->featureContext->getAdminPassword(),
			'GET',
			"/apps/testing/api/v1/getextension/$mimeTypeNeedle",
			null,
			$this->featureContext->getOcsApiVersion()
		);
		Assert::assertEquals("200", $response->getStatusCode());
		$responseXml = HttpRequestHelper::getResponseXml($response)->data[0];
		$extensions = \json_decode(\json_encode($responseXml), true)['element'];
		if (\is_array($extensions)) {
			return $extensions;
		}
		return [$extensions];
	}
}
