<?php

/**
 * ownCloud
 *
 * @author Artur Neumann <artur@jankaritech.com>
 * @copyright Copyright (c) 2017 Artur Neumann artur@jankaritech.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU AFFERO GENERAL PUBLIC LICENSE for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

namespace Page;

use Behat\Mink\Element\NodeElement;
use SensioLabs\Behat\PageObjectExtension\PageObject\Exception\ElementNotFoundException;

/**
 * Page Object for a Firewall Group
 * to keep the page objects small we do not define the whole page as one object
 * but logical parts of the page
 *
 */
class FileFirewallGroup extends OwncloudPage {
	/**
	 * @var NodeElement of this group
	 */
	protected $groupElement;

	protected $ruleContainerXpath = './/*[@class="rule-container"]';
	protected $groupNameInputFieldName = 'groupName';
	protected $addRuleBtnXpath = './/*[@data-add="rule"]';
	protected $deleteGroupBtnXpath = './/button[@data-delete="group"]';
	
	/**
	 * sets the NodeElement for the current group
	 * a little bit like __construct() but as we access this "sub-page-object"
	 * from an other Page Object by $this->getPage("FileFirewallGroup")
	 * there is no real __construct() that can take arguments
	 *
	 * @param NodeElement $groupElement
	 *
	 * @return void
	 */
	public function setElement(NodeElement $groupElement) {
		$this->groupElement = $groupElement;
	}

	/**
	 * returns the NodeElements for all rule containers
	 *
	 * @throws ElementNotFoundException
	 * @return \Behat\Mink\Element\NodeElement[]
	 */
	public function getAllRuleContainers() {
		$containers = $this->groupElement->findAll(
			"xpath", $this->ruleContainerXpath
		);
		if ($containers === null) {
			throw new ElementNotFoundException(
				"could not find any rules in this group"
			);
		}
		return $containers;
	}

	/**
	 * checks if a group is empty
	 *
	 * @return boolean
	 */
	public function isEmpty() {
		$isEmpty = true;
		foreach ($this->getAllRuleContainers() as $rule) {
			$ruleObject = $this->getPage("FileFirewallRule");
			$ruleObject->setElement($rule);
			if (!$ruleObject->isEmpty()) {
				$isEmpty = false;
				break;
			}
		}
		return $isEmpty;
	}

	/**
	 * sets the name of a group
	 *
	 * @param string $name
	 *
	 * @return void
	 */
	public function setName($name) {
		$this->groupElement->fillField($this->groupNameInputFieldName, $name);
	}

	/**
	 * deletes the current group
	 *
	 * @throws \Exception
	 * @return void
	 */
	public function delete() {
		$deleteGroupBtn = $this->groupElement->find(
			"xpath", $this->deleteGroupBtnXpath
		);
		if ($deleteGroupBtn === null) {
			throw new ElementNotFoundException("cannot find delete group button");
		}
		$deleteGroupBtn->click();
	}

	/**
	 * adds a new firewall rule to the current group
	 *
	 * @param string $filter
	 * @param string $operator
	 * @param string $value
	 * @param string $type (normal|regex)
	 * @param int $timeout_msec timeout in milliseconds
	 * @param bool $useEmptySpot if set to true and the last rule in the group
	 *                           is not set yet this empty rule slot will be
	 *                           used, otherwise a new slot will be added to
	 *                           the group
	 *
	 * @return FileFirewallRule
	 * @throws \Exception
	 * @throws \InvalidArgumentException
	 * @throws ElementNotFoundException
	 */
	public function addRule(
		$filter, $operator, $value, $type="normal",
		$timeout_msec=10000, $useEmptySpot = true
	) {
		if ($type !== "normal" && $type !== "regex") {
			throw new \InvalidArgumentException(
				"type can only be 'normal' or 'regex'"
			);
		}
		/**
		 * @var FileFirewallRule $ruleObject
		 * */
		$ruleObject = $this->getPage("FileFirewallRule");
		$allRules = $this->getAllRuleContainers();
		$oldRulesCount = \count($allRules);
		$ruleObject->setElement($allRules [$oldRulesCount - 1]);
		if ($useEmptySpot === false || !$ruleObject->isEmpty()) {
			$addRuleBtn = $this->groupElement->find("xpath", $this->addRuleBtnXpath);
			if ($addRuleBtn === null) {
				throw new ElementNotFoundException("cannot find add-rule button");
			}
			$addRuleBtn->click();

			$currentTime = \microtime(true);
			$end = $currentTime + ($timeout_msec / 1000);
			while ($currentTime <= $end) {
				$allRules = $this->getAllRuleContainers();
				if (\count($allRules) > $oldRulesCount) {
					break;
				}
				\usleep(STANDARD_SLEEP_TIME_MICROSEC);
				$currentTime = \microtime(true);
			}
			
			if (\count($allRules) <= $oldRulesCount) {
				throw new \Exception("could not add a new rule");
			}
		}
		$newRule = $allRules[\count($allRules) - 1];
		$ruleObject->setElement($newRule);
		
		try {
			$ruleObject->setType($type);
			$ruleObject->selectFilter($filter);
			$ruleObject->selectOperator($operator);
			$ruleObject->setValue($value);
		} catch (\Exception $e) {
			throw new \Exception("could not set rule settings", $e->getCode(), $e);
		}

		return $ruleObject;
	}

	/**
	 *
	 * @param string $filter
	 *
	 * @return void
	 * @throws \Exception
	 */
	public function findRuleByFilter($filter) {
		throw new \Exception("not implemented");
	}

	/**
	 *
	 * @param int $ruleNo
	 *
	 * @return \SensioLabs\Behat\PageObjectExtension\PageObject\Page
	 */
	public function getRuleByNo($ruleNo) {
		$ruleContainers = $this->getAllRuleContainers();
		$ruleObject = $this->getPage("FileFirewallRule");
		$ruleObject->setElement($ruleContainers[$ruleNo]);
		return $ruleObject;
	}
}
