<?php
/**
 * ownCloud Firewall
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright Copyright (c) 2018, ownCloud GmbH
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Firewall\Tests\Rules;

use OCA\Firewall\Ruler;
use OCA\Firewall\Rules\SystemTag;
use OCA\Firewall\Tests\Unit\Helpers\RuleTester;
use OCP\SystemTag\TagNotFoundException;

class SystemTagTest extends RuleTester {

	/** @var \OCP\SystemTag\ISystemTagObjectMapper|\PHPUnit\Framework\MockObject\MockObject */
	protected $tagMapper;

	/** @var \OCP\SystemTag\ISystemTagManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $tagManager;

	public function setUp(): void {
		parent::setUp();

		$this->tagManager = $this->getMockBuilder('OCP\SystemTag\ISystemTagManager')
			->disableOriginalConstructor()
			->getMock();

		$this->tagMapper = $this->getMockBuilder('OCP\SystemTag\ISystemTagObjectMapper')
			->disableOriginalConstructor()
			->getMock();

		$this->rule = new SystemTag(
			$this->ruler,
			$this->context,
			$this->tagManager,
			$this->tagMapper
		);
	}

	public function dataValidateRuleValue() {
		return [
			['0', false],
			['1', false],
			['100', false],
			['test', false],
			[0, false],
			[1, false],
			[42, false],
			[true, true],
			[false, true],
			[[], true],
			[['test'], true],
			[null, true],
		];
	}

	public function dataValidateRuleValueInvalidTags() {
		return [
			[new TagNotFoundException()],
			[new \InvalidArgumentException()],
		];
	}

	/**
	 * @dataProvider dataValidateRuleValueInvalidTags
	 *
	 * @param \Exception $toThrow
	 */
	public function testValidateRuleValueInvalidTags($toThrow) {
		$this->expectException(\OutOfBoundsException::class);

		$this->tagManager->expects($this->once())
			->method('getTagsByIds')
			->with('tag')
			->willThrowException($toThrow);

		$this->rule->validateRuleValue('tag', 'test');
	}

	/**
	 * @return string
	 */
	protected function getRuleClass() {
		return null;
	}

	public function dataDoCheck() {
		return [
			[1, [12], false, Ruler::OPERATOR_EQUALS, false],
			[2, [42], true, Ruler::OPERATOR_EQUALS, true],
			[3, [12, 21], new TagNotFoundException(), Ruler::OPERATOR_EQUALS, true],
			[4, [12, 21], false, Ruler::OPERATOR_NOT_EQUALS, true],
			[5, [42], true, Ruler::OPERATOR_NOT_EQUALS, false],
			[6, [12], new TagNotFoundException(), Ruler::OPERATOR_NOT_EQUALS, true],
		];
	}

	/**
	 * @dataProvider dataDoCheck
	 *
	 * @param int $ruleValue
	 * @param bool $contextValue
	 * @param mixed $mapperResponse
	 * @param string $operator
	 * @param bool $expected
	 */
	public function testDoCheck($ruleValue, $contextValue, $mapperResponse, $operator, $expected) {
		if (\is_bool($mapperResponse)) {
			$this->tagMapper->expects($this->once())
				->method('haveTag')
				->with($contextValue, 'files', $ruleValue, false)
				->willReturn($mapperResponse);
		} else {
			$this->tagMapper->expects($this->once())
				->method('haveTag')
				->willThrowException($mapperResponse);
		}
		$this->context->expects($this->once())
			->method('getFileIds')
			->willReturn($contextValue);

		$this->assertSame($expected, $this->rule->doCheck($operator, $ruleValue));
	}
}
