<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Controller;

use OCA\Workflow\Controller\RetentionController;
use OCA\Workflow\Retention\Exception\TagAlreadyHasRetention;
use OCA\Workflow\Retention\Exception\TagHasNoRetention;
use OCA\Workflow\Tests\Unit\TestCase;
use OCP\AppFramework\Http;
use OCP\SystemTag\TagNotFoundException;

class RetentionControllerTest extends TestCase {

	/** @var \OCA\Workflow\Retention\Manager|\PHPUnit\Framework\MockObject\MockObject */
	protected $manager;

	/** @var \OCP\SystemTag\ISystemTagManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $tagManager;

	public function setUp(): void {
		parent::setUp();

		$this->manager = $this->getMockBuilder('OCA\Workflow\Retention\Manager')
			->disableOriginalConstructor()
			->getMock();

		$this->tagManager = $this->getMockBuilder('OCP\SystemTag\ISystemTagManager')
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return RetentionController|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getController(array $methods = []) {
		/** @var $request \OCP\IRequest|\PHPUnit\Framework\MockObject\MockObject */
		$request = $this->getMockBuilder('OCP\IRequest')
			->disableOriginalConstructor()
			->getMock();

		/** @var $l \OCP\IL10N|\PHPUnit\Framework\MockObject\MockObject */
		$l = $this->getMockBuilder('OCP\IL10N')
			->disableOriginalConstructor()
			->getMock();
		$l->expects($this->any())
			->method('t')
			->willReturnArgument(0);

		if (empty($methods)) {
			return new RetentionController(
				'workflow',
				$request,
				$this->manager,
				$this->tagManager,
				$l
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\Controller\RetentionController')
				->setConstructorArgs([
					'workflow',
					$request,
					$this->manager,
					$this->tagManager,
					$l
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataGetRetentionPeriods() {
		return [
			[[]],
			[['not_empty']],
		];
	}

	/**
	 * @dataProvider dataGetRetentionPeriods
	 * @param array $data
	 */
	public function testGetRetentionPeriods($data) {
		$this->manager->expects($this->once())
			->method('getAll')
			->willReturn($data);

		$controller = $this->getController();
		$response = $controller->getRetentionPeriods();

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals(Http::STATUS_OK, $response->getStatus());
		$this->assertEquals($data, $response->getData());
	}

	public function dataAddRetentionPeriods() {
		return [
			['tag1', 10, 'days', null, ['retention-data1'], Http::STATUS_OK, ['retention-data1']],
			['tag2', 23, 'months', null, ['retention-data2'], Http::STATUS_OK, ['retention-data2']],
			['tag1', 10, 'days', new \OutOfBoundsException('validation error'), null, Http::STATUS_BAD_REQUEST, ['error' => 'validation error']],
			['tag1', 10, 'days', null, new TagAlreadyHasRetention(), Http::STATUS_BAD_REQUEST, ['error' => 'The tag already has a retention period']],
		];
	}

	/**
	 * @dataProvider dataAddRetentionPeriods
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 * @param mixed $validationThrows
	 * @param mixed $managerReturns
	 * @param int $expectedStatus
	 * @param array $expectedData
	 */
	public function testAddRetentionPeriod($tagId, $numUnits, $unit, $validationThrows, $managerReturns, $expectedStatus, $expectedData) {
		$controller = $this->getController(['validate']);
		if ($validationThrows instanceof \Exception) {
			$controller->expects($this->once())
				->method('validate')
				->with($tagId, $numUnits, $unit)
				->willThrowException($validationThrows);
		} else {
			$controller->expects($this->once())
				->method('validate')
				->with($tagId, $numUnits, $unit);
		}

		if ($managerReturns === null) {
			$this->manager->expects($this->never())
				->method('add');
		} elseif ($managerReturns instanceof \Exception) {
			$this->manager->expects($this->once())
				->method('add')
				->with($tagId, $numUnits, $unit)
				->willThrowException($managerReturns);
		} else {
			$this->manager->expects($this->once())
				->method('add')
				->with($tagId, $numUnits, $unit)
				->willReturn($managerReturns);
		}

		$response = $controller->addRetentionPeriod($tagId, $numUnits, $unit);

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals($expectedStatus, $response->getStatus());
		$this->assertEquals($expectedData, $response->getData());
	}

	public function dataUpdateRetentionPeriods() {
		return [
			['tag1', 10, 'days', null, ['retention-data1'], Http::STATUS_OK, ['retention-data1']],
			['tag2', 23, 'months', null, ['retention-data2'], Http::STATUS_OK, ['retention-data2']],
			['tag1', 10, 'days', new \OutOfBoundsException('validation error'), null, Http::STATUS_BAD_REQUEST, ['error' => 'validation error']],
			['tag1', 10, 'days', null, new TagHasNoRetention(), Http::STATUS_BAD_REQUEST, ['error' => 'The tag does not have a retention period']],
		];
	}

	/**
	 * @dataProvider dataUpdateRetentionPeriods
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 * @param mixed $validationThrows
	 * @param mixed $managerReturns
	 * @param int $expectedStatus
	 * @param array $expectedData
	 */
	public function testUpdateRetentionPeriod($tagId, $numUnits, $unit, $validationThrows, $managerReturns, $expectedStatus, $expectedData) {
		$controller = $this->getController(['validate']);
		if ($validationThrows instanceof \Exception) {
			$controller->expects($this->once())
				->method('validate')
				->with($tagId, $numUnits, $unit)
				->willThrowException($validationThrows);
		} else {
			$controller->expects($this->once())
				->method('validate')
				->with($tagId, $numUnits, $unit);
		}

		if ($managerReturns === null) {
			$this->manager->expects($this->never())
				->method('update');
		} elseif ($managerReturns instanceof \Exception) {
			$this->manager->expects($this->once())
				->method('update')
				->with($tagId, $numUnits, $unit)
				->willThrowException($managerReturns);
		} else {
			$this->manager->expects($this->once())
				->method('update')
				->with($tagId, $numUnits, $unit)
				->willReturn($managerReturns);
		}

		$response = $controller->updateRetentionPeriod($tagId, $numUnits, $unit);

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals($expectedStatus, $response->getStatus());
		$this->assertEquals($expectedData, $response->getData());
	}

	public function dataDeleteRetentionPeriods() {
		return [
			[21, null],
			[23, new TagHasNoRetention()],
		];
	}

	/**
	 * @dataProvider dataDeleteRetentionPeriods
	 * @param string $tagId
	 * @param mixed $managerReturns
	 */
	public function testDeleteRetentionPeriod($tagId, $managerReturns) {
		$controller = $this->getController();

		if ($managerReturns instanceof \Exception) {
			$this->manager->expects($this->once())
				->method('delete')
				->with($tagId)
				->willThrowException($managerReturns);
		} else {
			$this->manager->expects($this->once())
				->method('delete')
				->with($tagId);
		}

		$response = $controller->deleteRetentionPeriod($tagId);

		$this->assertInstanceOf('OCP\AppFramework\Http\Response', $response);
		$this->assertEquals(Http::STATUS_OK, $response->getStatus());
	}

	public function dataValidate() {
		return [
			['tag1', 23, 'days'],
			['tag2', 42, 'weeks'],
			['tag3', 21, 'months'],
			['tag4', 16, 'years'],
		];
	}

	/**
	 * @dataProvider dataValidate
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 */
	public function testValidate($tagId, $numUnits, $unit) {
		$controller = $this->getController();

		$this->tagManager->expects($this->once())
			->method('getTagsByIds')
			->with($tagId);

		$this->invokePrivate($controller, 'validate', [$tagId, $numUnits, $unit]);
		$this->assertTrue(true);
	}

	public function dataValidateInvalid() {
		return [
			['doesnotexist', 42, 'days', 'The given tag does not exist', 1],

			['tag1', 0, 'days', 'Invalid retention period', 2],
			['tag1', -1, 'days', 'Invalid retention period', 2],

			['tag1', 42, 'hours', 'Invalid retention period unit', 3],
			['tag1', 42, 'minutes', 'Invalid retention period unit', 3],
			['tag1', 42, 'seconds', 'Invalid retention period unit', 3],
			['tag1', 42, 'y', 'Invalid retention period unit', 3],
		];
	}

	/**
	 * @dataProvider dataValidateInvalid
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 * @param string $exception
	 * @param int $code
	 */
	public function testValidateInvalid($tagId, $numUnits, $unit, $exception, $code) {
		$this->expectException('OutOfBoundsException', $exception, $code);

		if ($tagId === 'doesnotexist') {
			$this->tagManager->expects($this->once())
				->method('getTagsByIds')
				->with($tagId)
				->willThrowException(new TagNotFoundException());
		} else {
			$this->tagManager->expects($this->once())
				->method('getTagsByIds')
				->with($tagId);
		}

		$controller = $this->getController();
		$this->invokePrivate($controller, 'validate', [$tagId, $numUnits, $unit]);
		$this->assertTrue(true);
	}
}
