<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Engine;

use OCA\Workflow\Engine\Exception\FlowDoesNotExist;
use OCA\Workflow\Engine\Flow;
use OCA\Workflow\Engine\FlowManager;
use OCA\Workflow\Tests\Unit\TestCase;

/**
 * Class FlowManagerTest
 *
 * @group DB
 * @package OCA\Workflow\Tests\Flow
 */
class FlowManagerTest extends TestCase {

	/** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface|\PHPUnit\Framework\MockObject\MockObject */
	protected $dispatcher;

	public function setUp(): void {
		parent::setUp();

		$this->dispatcher = $this->getMockBuilder('Symfony\Component\EventDispatcher\EventDispatcherInterface')
			->disableOriginalConstructor()
			->getMock();

		$this->resetTable();
	}

	/**
	 * @param array $methods
	 * @return FlowManager|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getManager(array $methods = []) {
		if (empty($methods)) {
			return new FlowManager(
				\OC::$server->getDatabaseConnection(),
				$this->dispatcher
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\Engine\FlowManager')
				->setConstructorArgs([
					\OC::$server->getDatabaseConnection(),
					$this->dispatcher,
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function tearDown(): void {
		$this->resetTable();

		parent::tearDown();
	}

	protected function resetTable() {
		$query = \OC::$server->getDatabaseConnection()->getQueryBuilder();

		$query->delete('workflows')
			->execute();
	}

	public function dataGetFlows() {
		return [
			[[]],
			[['not_empty']],
		];
	}

	/**
	 * @dataProvider dataGetFlows
	 * @param array $getFlows
	 */
	public function testGetRetentionFlows(array $getFlows) {
		$manager = $this->getManager(['getFlows']);

		$manager->expects($this->once())
			->method('getFlows')
			->with('retention')
			->willReturn($getFlows);

		$this->assertSame($getFlows, $manager->getRetentionFlows());
	}

	/**
	 * @dataProvider dataGetFlows
	 * @param array $getFlows
	 */
	public function testGetAutoTaggingFlows(array $getFlows) {
		$manager = $this->getManager(['getFlows', 'getTypes']);

		$manager->expects($this->once())
			->method('getTypes')
			->willReturn(['workflow_autotagging' => ['description' => 'foo', 'supportedFileActions' => []]]);
		$manager->expects($this->once())
			->method('getFlows')
			->with(['workflow_autotagging'])
			->willReturn($getFlows);

		$this->assertSame($getFlows, $manager->getFlowsForTrigger(null));
	}

	public function testFull() {
		$manager = $this->getManager();
		$this->assertCount(0, $manager->getRetentionFlows());
		$this->assertCount(0, $manager->getFlowsForTrigger(null));

		// Add
		$flow1 = $this->getFlow(1);
		$flowId = $manager->addFlow($flow1);
		$this->assertGreaterThan(0, $flowId);

		// GetAll
		$flow1 = $this->getFlow(1, $flowId);

		$this->assertCount(1, $manager->getRetentionFlows());
		$this->assertCount(0, $manager->getFlowsForTrigger(null));
		$all = $this->invokePrivate($manager, 'getFlows', [null]);
		$this->assertCount(1, $all);
		$this->assertSameFlow($flow1, $all[0], true);

		// Get
		$result = $manager->getFlowById($flowId);
		$this->assertSameFlow($flow1, $result);

		// Update
		$flow2 = $this->getFlow(2, $flowId);
		$this->assertTrue($manager->updateFlow($flow2));
		$result = $manager->getFlowById($flowId);
		$this->assertSameFlow($flow2, $result);
		$this->assertCount(1, $this->invokePrivate($manager, 'getFlows', [null]));

		// Delete
		$manager->deleteFlow($result);
		$this->assertEmpty($this->invokePrivate($manager, 'getFlows', [null]));

		// Get non existing
		try {
			$manager->getFlowById($flowId);
			$this->fail('Expected FlowDoesNotExist to be thrown');
		} catch (FlowDoesNotExist $e) {
			$this->assertTrue(true);
		}
		$this->assertCount(0, $this->invokePrivate($manager, 'getFlows', [null]));
	}

	public function dataFlowToArray() {
		return [
			[$this->getFlow(1, 23), ['id' => 23, 'type' => 'retention', 'name' => 'name1', 'conditions' => ['conds1'], 'actions' => ['actions1']]],
			[$this->getFlow(2, 42), ['id' => 42, 'type' => 'retention', 'name' => 'name2', 'conditions' => ['conds2'], 'actions' => ['actions2']]],
		];
	}

	/**
	 * @dataProvider dataFlowToArray
	 *
	 * @param Flow $flow
	 * @param array $expected
	 */
	public function testFlowToArray(Flow $flow, array $expected) {
		$manager = $this->getManager();
		$this->assertEquals($expected, $manager->flowToArray($flow));
	}

	protected function assertSameFlow(Flow $flow1, Flow $flow2, $compareId = false) {
		$this->assertSame($flow1->getName(), $flow2->getName());
		$this->assertSame($flow1->getType(), $flow2->getType());
		$this->assertSame($flow1->getConditions(), $flow2->getConditions());
		$this->assertSame($flow1->getActions(), $flow2->getActions());

		if ($compareId) {
			$this->assertSame($flow1->getId(), $flow2->getId());
		}
	}

	/**
	 * @param int $int
	 * @param int $id
	 * @return \OCA\Workflow\Engine\Flow
	 */
	protected function getFlow($int, $id = 0) {
		$flow = $this->getMockBuilder('OCA\Workflow\Engine\Flow')
			->disableOriginalConstructor()
			->getMock();
		$flow->expects($this->atLeastOnce())
			->method('getName')
			->willReturn('name' . $int);
		$flow->expects($this->atLeastOnce())
			->method('getType')
			->willReturn('retention');
		$flow->expects($this->atLeastOnce())
			->method('getConditions')
			->willReturn(['conds' . $int]);
		$flow->expects($this->atLeastOnce())
			->method('getActions')
			->willReturn(['actions' . $int]);

		if ($id) {
			$flow->expects($this->atLeastOnce())
				->method('getId')
				->willReturn($id);
		}

		return $flow;
	}
}
